from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import NDArray
from sklearn.manifold import TSNE


def get_tsne_clusters(clusters: NDArray,
                      labels: NDArray,
                      centroids: Optional[NDArray] = None):

    fig = plt.figure()
    axe = fig.add_subplot(1, 1, 1)
    color_map = {0: 'red', 1: 'green', 2: 'blue', 3: 'yellow', 4: 'purple'}

    colors = [color_map[label] for label in labels]
    tsne = TSNE(n_components=2, random_state=42)

    if not isinstance(centroids, type(None)):
        concatenated = np.concatenate((clusters, np.array(centroids)), axis=0)
        clusters_tsne = tsne.fit_transform(concatenated)

        axe.scatter(clusters_tsne[:-len(centroids), 0],
                    clusters_tsne[:-len(centroids), 1], c=colors)
        axe.scatter(clusters_tsne[-len(centroids):, 0],
                    clusters_tsne[-len(centroids):, 1], marker='*', s=100, c='black')
    else:
        clusters_tsne = tsne.fit_transform(clusters)
        axe.scatter(clusters_tsne[:, 0], clusters_tsne[:, 1], c=colors)
    return fig



import numpy as np
from numpy.typing import NDArray


def minkowski_distance(point_a: NDArray, point_b: NDArray, p: float) -> NDArray:
    '''
    Minkowski distance function.
    '''
    return np.power(np.sum(np.power(np.abs(point_a - point_b), p)), 1/p)


def pairwise_minkowski_distance(point_a: NDArray,
                                points: NDArray | list,
                                p: float
                                ) -> NDArray:
    '''
    Pairwise Minkowski distance function.
    '''

    result = np.array(
        [minkowski_distance(point_a, point, p) for point in points]
    )
    return result



import numpy as np
import pytest

from lkmeans import LKMeans

p_values = [0.5, 1, 2, 5]

@pytest.mark.api
@pytest.mark.parametrize("p", p_values)
def test_segment_slsqp_calculation(p) -> None:
    data = np.random.uniform(-10,10, size=(100, 50))

    lkmeans = LKMeans(n_clusters=2, p=p)
    lkmeans.fit_predict(data)
    print('Inertia', lkmeans.inertia_)
    print('Centers', lkmeans.cluster_centers_)



from argparse import ArgumentParser
from pathlib import Path

from lkmeans.examples.experiment import run_experiment
from lkmeans.examples.experiment_data import get_experiment_data

parser = ArgumentParser()

parser.add_argument(
    '--path',
    type=Path,
    default=Path('experiments'),
    help='Path to save results'
)

parser.add_argument(
    '--num-clusters',
    type=int,
    default=2
)


def main():
    args = parser.parse_args()
    experiments_path = args.path

    minkowski_parameter = [0.2, 0.6, 1, 1.5, 2, 3, 5]
    T_parameter = [0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    repeats = 100
    n_points = [100, 500, 1000]

    dimension = 20
    n_clusters, prob, mu_list, cov_matrices = get_experiment_data(
        num_clusters=args.num_clusters, dimension=dimension)

    for points in n_points:
        experiment_name = f'Clusters:{n_clusters}, points:{points}'
        output_path = experiments_path / f'exp_{args.num_clusters}_points_{points}'

        run_experiment(
            n_clusters=n_clusters,
            distance_parameters=T_parameter,
            minkowski_parameters=minkowski_parameter,
            repeats=repeats,
            n_points=points,
            cov_matrices=cov_matrices,
            prob=prob,
            mu_list=mu_list,
            experiment_name=experiment_name,
            output_path=output_path
        )


if __name__ == '__main__':
    main()



from argparse import ArgumentParser
from pathlib import Path

import matplotlib.pyplot as plt

from lkmeans.data_generation.points_generator import generate_mix_distribution
from lkmeans.examples.experiment_data import get_experiment_data

parser = ArgumentParser()

parser.add_argument(
    '--path',
    type=Path,
    default=Path('images'),
    help='Path to save results'
)


def main():
    args = parser.parse_args()
    args.path.mkdir(exist_ok=True)

    dimension = 20
    n_points = 100

    n_clusters, prob, mu_list, cov_matrices = get_experiment_data(num_clusters=2, dimension=dimension)

    for t in [0.2, 0.4, 0.9]:
        filename = args.path / f'{n_clusters}_cluster_hist_t_{t}.png'
        clusters, _, _ = generate_mix_distribution(
            probability=prob,
            mu_list=mu_list,
            cov_matrices=cov_matrices,
            n_samples=n_points,
            t=t
        )

        fig, ax = plt.subplots(figsize=(5, 3))
        ax.hist(clusters[:, 0], bins=15)
        ax.grid(True, color='gray', linestyle='--', linewidth=0.5)
        fig.savefig(str(filename), dpi=300, bbox_inches='tight')
        plt.close(fig)


if __name__ == '__main__':
    main()



import numpy as np
from numpy.typing import NDArray

from lkmeans.optimizers import (mean_optimizer, median_optimizer,
                                segment_slsqp_optimizer)


def get_test_data(size: int) -> tuple[NDArray, float]:
    centre = +89.9573

    data = np.random.random(size)
    reverted_data = data * -1
    samples = np.concatenate([data, reverted_data])
    samples = samples + centre
    return samples, centre


def main() -> None:
    samples, centre = get_test_data(50)

    print(f'Expected centre: {centre :.5f}')

    print(f'Optimizer median: {median_optimizer(samples) :.5f}')
    print(f'Optimizer mean: {mean_optimizer(samples) :.5f}')

    for p in [0.2, 0.5]:
        print(
            f'Optimizer SLSQP (p={p}): {segment_slsqp_optimizer(samples, p)}')


if __name__ == '__main__':
    main()



import numpy as np
import pytest
from numpy.typing import NDArray

from lkmeans.optimizers import (bound_optimizer, mean_optimizer,
                                median_optimizer, segment_slsqp_optimizer)


def get_test_data(size: int, center: float) -> NDArray:
    data = np.random.random(size)
    reverted_data = data * -1
    samples = np.concatenate([data, reverted_data])
    samples = samples + center
    return samples


testing_params = [-356.189, -2.56, 0, 1.111, 34.954]

params = [
    (np.array([1, 2, 3]), 2),
    (np.array([1, 2, 3, 4]), 2.5),
    (np.array([0]), 0),
    (np.array([0, 0]), 0),
    (np.array([1, 3, 5, 7]), 4),
]

p_values = [0.01, 0.2, 0.5]


@pytest.mark.optimizers
@pytest.mark.parametrize("median", testing_params)
def test_median_calculation(median) -> None:
    n_samples = 50
    samples = get_test_data(n_samples, median)
    np.testing.assert_almost_equal(median, median_optimizer(samples))


@pytest.mark.optimizers
@pytest.mark.parametrize("test_input, expected_output", params)
def test_median_optimizer_inputs(test_input, expected_output):
    np.testing.assert_almost_equal(mean_optimizer(test_input), expected_output)


@pytest.mark.optimizers
@pytest.mark.parametrize("median", testing_params)
def test_mean_optimizer(median) -> None:
    n_samples = 50
    samples = get_test_data(n_samples, median)
    np.testing.assert_almost_equal(median, mean_optimizer(samples))


@pytest.mark.optimizers
@pytest.mark.parametrize("test_input, expected_output", params)
def test_mean_optimizer_inputs(test_input, expected_output):
    np.testing.assert_almost_equal(mean_optimizer(test_input), expected_output)


@pytest.mark.optimizers
@pytest.mark.parametrize("median", testing_params)
@pytest.mark.parametrize("p", p_values)
def test_segment_slsqp_calculation(median, p) -> None:
    n_samples = 50
    samples = get_test_data(n_samples, median)
    np.testing.assert_almost_equal(median, segment_slsqp_optimizer(samples, p))


@pytest.mark.optimizers
@pytest.mark.parametrize("test_input, expected_output", params)
def test_segment_slsqp_optimizer_inputs(test_input, expected_output):
    np.testing.assert_almost_equal(mean_optimizer(test_input), expected_output)


@pytest.mark.optimizers
@pytest.mark.parametrize("median", testing_params)
@pytest.mark.parametrize("p", p_values)
def test_bound_calculation(median, p) -> None:
    n_samples = 50
    samples = get_test_data(n_samples, median)
    np.testing.assert_almost_equal(median, bound_optimizer(samples, p), decimal=0)



from argparse import ArgumentParser
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np

from lkmeans.clustering import assign_to_cluster
from lkmeans.data_generation.points_generator import generate_mix_distribution
from lkmeans.distance import pairwise_minkowski_distance
from lkmeans.examples.experiment_data import get_experiment_data

parser = ArgumentParser()

parser.add_argument(
    '--path',
    type=Path,
    default=Path('images'),
    help='Path to save results'
)

parser.add_argument(
    '--p',
    type=float,
    default=2,
    help='Minkowski parameter'
)

parser.add_argument(
    '--t',
    type=float,
    default=0.,
    help='T parameter of distribution'
)


# pylint: disable=too-many-locals
def main():
    args = parser.parse_args()
    args.path.mkdir(exist_ok=True)
    p = int(args.p) if (args.p).is_integer() else args.p

    dimension = 20
    n_points = 10
    n_observation = 10000

    n_clusters, prob, mu_list, cov_matrices = get_experiment_data(num_clusters=2, dimension=dimension)

    filename = args.path / f'plot_minkowski_function_with_p_{p}.png'
    samples, _, centroids = generate_mix_distribution(
        probability=prob,
        mu_list=mu_list,
        cov_matrices=cov_matrices,
        n_samples=n_points,
        t=0.1
    )

    dim = 0

    clusters, _ = assign_to_cluster(samples, centroids, n_clusters, p)
    cluster = np.array(clusters[0])
    dimension_data = cluster[:,dim]

    points = np.linspace(min(dimension_data), max(dimension_data), n_observation)
    minkowski_values = pairwise_minkowski_distance(
        point_a = dimension_data,
        points=points,
        p=p
    )

    fig, ax = plt.subplots(figsize=(5, 3))
    ax.scatter(points, minkowski_values)
    ax.axis('off')
    fig.savefig(str(filename), dpi=300, bbox_inches='tight')
    plt.close(fig)


if __name__ == '__main__':
    main()



import time
from pathlib import Path

import numpy as np
from matplotlib import pyplot as plt
from numpy.typing import NDArray
from sklearn.metrics import adjusted_mutual_info_score, adjusted_rand_score

from lkmeans import LKMeans
from lkmeans.data_generation.points_generator import generate_mix_distribution
from lkmeans.report.decomposition import get_tsne_clusters
from lkmeans.report.metric_meter import (GraphicMeter, MetricMeter,
                                         MetricTable, insert_hline)


def get_covariance_matrix(sigma: float, dimension: int) -> NDArray:
    return np.eye(dimension) * sigma


# pylint: disable=too-many-arguments, too-many-locals
def repeat_iteration(
        repeats: int,
        n_clusters: int,
        n_points: int,
        prob: float,
        cov_matrices: list[NDArray],
        t: float,
        mu_list: list[NDArray],
        p: float | int,
        makes_plot: bool,
        output_path: Path

):
    repeat_metric_meter = MetricMeter()
    for _ in range(repeats):

        clusters, labels, centroids = generate_mix_distribution(
            probability=prob,
            mu_list=mu_list,
            cov_matrices=cov_matrices,
            n_samples=n_points,
            t=t
        )

        experiment_time = time.perf_counter()
        lkmeans = LKMeans(n_clusters=n_clusters, p=p)
        generated_labels = lkmeans.fit_predict(clusters)
        experiment_time = time.perf_counter() - experiment_time

        repeat_metric_meter.add_combination(
            ari=adjusted_rand_score(labels, generated_labels),
            ami=float(adjusted_mutual_info_score(labels, generated_labels)),
            inertia=lkmeans.inertia_,
            time=experiment_time
        )
        if makes_plot:
            figure_name = f'factor_{t:.1f}'.replace('.', '_')
            fig = get_tsne_clusters(clusters, labels, centroids)
            fig.savefig(
                str(output_path / f'{figure_name}.png'), dpi=300, bbox_inches='tight')
            plt.close(fig)
    average_ari, average_ami, average_inertia, average_time = repeat_metric_meter.get_average()
    return average_ari, average_ami, average_inertia, average_time


# pylint: disable= too-many-arguments, too-many-locals
def run_experiment(
        n_clusters: int,
        distance_parameters: list[float],
        minkowski_parameters: list[float | int],
        repeats: int,
        n_points: int,
        cov_matrices: list[NDArray],
        prob: float,
        mu_list: list[NDArray],
        experiment_name: str,
        output_path: Path,
        makes_plot: bool = False) -> None:
    '''Function for evaluation experiment'''

    output_path.mkdir(exist_ok=True, parents=True)

    table = MetricTable()

    generator = [GraphicMeter(distance_parameters, 't')
                 for _ in minkowski_parameters]
    graphic_t_metrics_dict = dict(zip(minkowski_parameters, generator))

    for t in distance_parameters:

        graphic_p_metrics = GraphicMeter(minkowski_parameters, 'p')
        for p in minkowski_parameters:

            average_ari, average_ami, average_inertia, average_time = repeat_iteration(
                repeats, n_clusters, n_points, prob,
                cov_matrices, t, mu_list, p, makes_plot, output_path
            )

            table.add_to_frame(
                ari=average_ari,
                ami=average_ami,
                inertia=average_inertia,
                time=average_time,
                name=f'{experiment_name}, T:{t:.1f}, P:{p}'
            )

            graphic_p_metrics.add_combination(
                ari=average_ari,
                ami=average_ami,
                inertia=average_inertia,
                time=average_time
            )

            graphic_t_metrics_dict[p].add_combination(
                ari=average_ari,
                ami=average_ami,
                inertia=average_inertia,
                time=average_time
            )

        for metric_graph in ['ARI', 'AMI', 'Inertia', 'Time']:
            figure_name = f'factor_{t:.1f}_{metric_graph}'.replace('.', '_')
            fig = graphic_p_metrics.get_graph(metric_graph)
            fig.savefig(
                str(output_path / f'{figure_name}.png'), dpi=300, bbox_inches='tight')
            plt.close(fig)

    print(table.get_table())

    for p, graph_t_meter in graphic_t_metrics_dict.items():
        for metric in ['ARI', 'AMI', 'Inertia', 'Time']:
            figure_name = f'{metric}_by_t_with_p_{p}'.replace('.', '_')
            fig = graph_t_meter.get_graph(metric)
            fig.savefig(
                str(output_path / f'{figure_name}.png'), dpi=300, bbox_inches='tight')
            plt.close(fig)

    table_name = 'experiment 1'
    table = table.get_latex_table(caption='Experiment 1')
    table = insert_hline(table)

    latex_logs = output_path / f'{table_name.replace(" ", "_")}.tex'
    with latex_logs.open('w') as f:
        f.write(table)



import numpy as np
from numpy.typing import NDArray


def move_towards_mean(mu: NDArray, mu_mean: NDArray, t: float) -> NDArray:
    '''
    Returns a new point that is moved towards the mean according to the rule:
    point(t) = point + t * (mean - point), 0 ≤ t ≤ 1
    '''
    new_mu = mu + t * (mu_mean - mu)
    return new_mu


def split_points_by_clusters(n_points: int, n_clusters: int, probability: float) -> list[int]:
    '''
    Get the number of points to generate and the number of clusters as input
    returns the list with number of point in each cluster
    '''
    n_list = []
    if n_clusters == 2:
        n_1 = int(probability * n_points)
        n_list = [n_1, n_points - n_1]
    elif n_clusters == 3:
        n_1 = int(probability * n_points)
        n_2 = int(probability * n_points)
        n_3 = n_points - n_1 - n_2
        n_list = [n_1, n_2, n_3]
    return n_list


# pylint: disable= too-many-locals
def generate_mix_distribution(
    probability: float,
    mu_list: list[NDArray],
    cov_matrices: list[NDArray],
    n_samples: int,
    t: float
    ) -> tuple[NDArray, NDArray, NDArray]:
    '''
    Returns points from gaussian distributions.
    generated by distribution probability, covariance matrix and means vector.
    '''
    n_clusters = len(mu_list)

    mu_mean = np.mean(mu_list, axis=0)
    modified_mu_list = [move_towards_mean(mu, mu_mean, t) for mu in mu_list]

    splitted_points = split_points_by_clusters(n_samples, n_clusters, probability)

    distributions = []
    for n, mu, covariance_matrix in zip(splitted_points, modified_mu_list, cov_matrices):
        distribution = np.random.multivariate_normal(
            np.squeeze(mu, axis=0), covariance_matrix, n)
        distributions.append(distribution)

    samples = np.concatenate(distributions, axis=0)

    labels = np.array([])
    for label_id, n in enumerate(splitted_points):
        samples_labels = np.full(n, label_id)
        labels = np.concatenate((labels, samples_labels))

    # Shuffle the samples and labels
    permutation = np.random.permutation(n_samples)
    samples = samples[permutation]
    labels = labels[permutation]

    return samples, labels, np.mean(modified_mu_list, axis=1)



import numpy as np
from sklearn import metrics
from sklearn.cluster import KMeans

from lkmeans import LKMeans


def rgb2gray(rgb):
    return np.dot(rgb[...,:3], [0.2989, 0.5870, 0.1140])


def preprocess_cifar(data: np.ndarray) -> np.ndarray:
    data = rgb2gray(data)
    data = data.astype('float32')
    data = data.reshape(len(data),-1)
    return data


def preprocess_mnist(data: np.ndarray) -> np.ndarray:
    data = data.astype('float32')
    data = data/255.0
    data = data.reshape(len(data),-1)
    return data


def get_metrics(true_labels, predicted_labels, inertia: float):
    ami = metrics.adjusted_mutual_info_score(true_labels, predicted_labels)
    ari = metrics.adjusted_rand_score(true_labels, predicted_labels)
    return {'ARI': ari, 'AMI': ami, 'Inertia': inertia}


def make_experiment(data, labels):
    (n_samples, n_features), n_digits = data.shape, np.unique(labels).size
    print(f"# classes: {n_digits}; # samples: {n_samples}; # features {n_features}")


    kmeans = KMeans(init='random', n_clusters=n_digits, n_init=1, random_state=0)
    result_labels = kmeans.fit_predict(data)
    print('k-Means (sklearn)', get_metrics(labels, result_labels, kmeans.inertia_))


    for p in [2, 5, 1, 0.5]:
        lkmeans = LKMeans(n_clusters=n_digits, p=p)
        result_labels = lkmeans.fit_predict(data)
        print(f'LKMeans (p={p})', get_metrics(labels, result_labels, lkmeans.inertia_))



import warnings
from functools import partial
from multiprocessing import Pool, cpu_count

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import minimize

from lkmeans.distance import minkowski_distance


def median_optimizer(dimension_slice: NDArray) -> float:
    '''
    Standard K-Medoids optimizer.
    Returns the optimal `median` point of input array
    '''
    return float(np.median(dimension_slice))


def mean_optimizer(dimension_slice: NDArray) -> float:
    '''
    Standard K-Means optimizer.
    Returns the optimal `mean` point of input array
    '''
    return float(np.mean(dimension_slice))


def bound_optimizer(dimension_slice: NDArray, p: float | int) -> float:
    '''
    Special LKMeans optimizer.
    Based on idea that for 0 < p < 1 the minkowski function is a concave function.
    Returns the optimal point of input array with Minkowski metric parameter `p`
    '''
    points = np.unique(dimension_slice)

    optimal_point = points[0]
    smallest_distant = minkowski_distance(optimal_point, dimension_slice, p)
    for applicant in points:
        distance_of_applicant = minkowski_distance(applicant, dimension_slice, p)
        if distance_of_applicant < smallest_distant:
            optimal_point = applicant
            smallest_distant = distance_of_applicant
    return float(optimal_point)


def parallel_segment_slsqp_optimizer(cluster: NDArray, dim: int, p: float | int):
    optimize_slice = partial(segment_slsqp_optimizer, p=p)
    dimension_slices = [cluster[:, coordinate_id] for coordinate_id in range(dim)]
    with Pool(cpu_count()) as pool:
        new_centroid = pool.map(optimize_slice, dimension_slices)
    return new_centroid


def segment_slsqp_optimizer(dimension_slice: NDArray, p: float | int, tol: float = 1e-1_000) -> float:
    dimension_slice = np.unique(dimension_slice)

    median = np.median(dimension_slice)
    fun_median = minkowski_distance(
        np.array(median), dimension_slice, p)

    minimized_fun_median = fun_median
    for bound_id in range(len(dimension_slice) - 1):

        bounds = [(dimension_slice[bound_id],
                   dimension_slice[bound_id + 1])]

        x0 = np.mean(bounds)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            res = minimize(
                fun=lambda centre: minkowski_distance(
                    centre, dimension_slice, p),
                x0=x0,
                method='SLSQP',
                bounds=bounds,
                tol=tol
            )

        if res.success:
            minima_point = res.x[0]
            minimal_point_value = res.fun
            if minimal_point_value < minimized_fun_median:
                minimized_fun_median = minimal_point_value
                median = minima_point
    return float(median)


def slsqp_optimizer(dimension_slice: NDArray, p: float | int, tol: float = 1e-1_000):
    x0 = np.mean(dimension_slice)
    bounds = [(min(dimension_slice), max(dimension_slice))]

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        res = minimize(
            fun=lambda centre: minkowski_distance(
                centre, dimension_slice, p),
            x0=x0,
            method='SLSQP',
            bounds=bounds,
            tol=tol
        ).x[0]
    return float(res)



import numpy as np
import pytest
from numpy.typing import NDArray

from lkmeans.clustering import assign_to_cluster


def _get_test_params():
    params = [
        (
            np.array([
                [0, 0], [0, 0], [0, 0], [0, 0], [0, 0],
                [1, 1], [1, 1], [1, 1], [1, 1], [1, 1]]
            ), np.array([[0, 0], [1, 1]]),
            np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
        ),
        (
            np.array([
                [0.500001, 0.50000001], [0, 0], [0, 0], [0, 0], [0, 0],
                [1, 1], [1, 1], [1, 1], [1, 1], [1, 1]]
            ), np.array([[0, 0], [1, 1]]),
            np.array([1, 0, 0, 0, 0, 1, 1, 1, 1, 1])
        )
    ]
    return params


def _get_test_p() -> list[float | int]:
    p = [0.2, 0.5, 1, 2]
    return p


@pytest.mark.lkmeans
@pytest.mark.parametrize("test_input, centroids, expected_output", _get_test_params())
@pytest.mark.parametrize("p", _get_test_p())
def test_assign_to_cluster(test_input: NDArray,
                           centroids: NDArray,
                           expected_output: NDArray,
                           p: float | int) -> None:
    _, label = assign_to_cluster(test_input, centroids, len(centroids), p)
    np.testing.assert_array_equal(expected_output, label)



'''Evaluation of LKMeans algorithm on real datasets'''

import keras.datasets
import sklearn.datasets

from lkmeans.examples.utils import make_experiment, preprocess_mnist


def main() -> None:
    # # Digits
    # print('='*50, '\nDigits dataset results')
    # data, labels = sklearn.datasets.load_digits(return_X_y=True)
    # make_experiment(data, labels)


    # Wine
    print('='*50, '\nWine dataset results')
    data, labels = sklearn.datasets.load_wine(return_X_y=True)
    make_experiment(data, labels)


    # Breast Cancer
    print('='*50, '\nBreast Cancer dataset results')
    data, labels = sklearn.datasets.load_breast_cancer(return_X_y=True)
    make_experiment(data, labels)


    # MNIST
    print('='*50, '\nMNIST dataset results')
    (data, labels), (_, _) = keras.datasets.mnist.load_data()
    data = preprocess_mnist(data)
    make_experiment(data, labels)


    # Fashion-MNIST
    print('='*50, '\nFashion-MNIST dataset results')
    (data, labels), (_, _) = keras.datasets.fashion_mnist.load_data()
    data = preprocess_mnist(data)
    make_experiment(data, labels)


    # # CIFAR10
    # print('='*50, '\nCIFAR10 dataset results')
    # (data, labels), (_, _) = keras.datasets.cifar10.load_data()
    # data = preprocess_cifar(data)
    # labels = labels.squeeze()
    # make_experiment(data, labels)


    # # CIFAR100
    # print('='*50, '\nCIFAR100 dataset results')
    # (data, labels), (_, _) = keras.datasets.cifar100.load_data()
    # data = preprocess_cifar(data)
    # labels = labels.squeeze()
    # make_experiment(data, labels)



if __name__ == '__main__':
    main()



from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


class MetricTable:
    def __init__(self) -> None:
        self.frames = []

    def add_frame(self, frame: pd.DataFrame) -> None:
        self.frames.append(frame)

    # pylint: disable=too-many-arguments
    def add_to_frame(self,
                     ari: float,
                     ami: float,
                     inertia: float,
                     time: float,
                     name: Optional[str] = 'Experiment'
                     ) -> None:
        data = {'ARI': f'{ari:.2f}', 'AMI': f'{ami:.2f}',
                'Inertia': f'{inertia:.2f}', 'Time': f'{time:.2f}'}
        frame = pd.DataFrame(data, [name])
        self.frames.append(frame)

    def add_empty_frame(self, time: bool) -> None:
        empty = 'N/A'
        data = {'ARI': empty, 'AMI': empty}
        if time:
            data['Time'] = empty
        frame = pd.DataFrame(data, [empty])
        self.frames.append(frame)

    def get_table(self) -> pd.DataFrame:
        return pd.concat(self.frames, join="inner")

    def get_latex_table(self, caption: str = '') -> str:
        table = self.get_table()
        return table.to_latex(index=True, escape=True, caption=caption)


def insert_hline(latex_str: str) -> str:
    lines_strings = latex_str.splitlines()
    result = []

    for line in lines_strings:
        if 'N/A' in line:
            result.append('\\midrule')
        else:
            result.append(line)
    result = '\n'.join(result)
    return result


class MetricMeter:
    def __init__(self) -> None:
        self.ari = []
        self.ami = []
        self.inertia = []
        self.time = []

    def add_ari(self, value: float) -> None:
        self.ari.append(value)

    def add_ami(self, value: float) -> None:
        self.ami.append(value)

    def add_inertia(self, value: float) -> None:
        self.inertia.append(value)

    def add_time(self, value: float) -> None:
        self.time.append(value)

    def add_combination(self, ari: float, ami: float, inertia: float, time: float) -> None:
        self.add_ari(abs(ari))
        self.add_ami(abs(ami))
        self.add_inertia(inertia)
        self.add_time(time)

    def get_average(self) -> tuple[float, float, float, float]:
        return float(np.mean(self.ari)), float(np.mean(self.ami)), \
            float(np.mean(self.inertia)), float(np.mean(self.time))


class GraphicMeter(MetricMeter):
    def __init__(self, base: list, base_name: str) -> None:
        super().__init__()
        self.base = base
        self.base_name = base_name

    def get_graph(self, key: str):
        values = {'ARI': self.ari, 'AMI': self.ami,
                  'Inertia': self.inertia, 'Time': self.time}

        fig, ax = plt.subplots(figsize=(5, 4))
        param = values[key]
        ax.plot(self.base, param, '-o')
        ax.grid(True, color='gray', linestyle='--', linewidth=0.5)

        if self.base_name == 'p':
            ax.set_xticks(self.base)
        else:
            ax.set_xticks(np.linspace(0, 1, 11))
        ax.set_xlabel(self.base_name)

        if key in ('ARI', 'AMI'):
            ax.set_yticks(np.arange(0, 1.1, 0.1))
        else:
            ax.set_yticks(np.linspace(np.min(param), np.max(param), 10))
        ax.set_ylabel(key)
        # ax.set_title(f'{key} vs. {self.base_name}')
        return fig



from copy import deepcopy

import numpy as np
from numpy.typing import NDArray

from lkmeans.distance import pairwise_minkowski_distance
from lkmeans.optimizers import (bound_optimizer, mean_optimizer,
                                median_optimizer, slsqp_optimizer)


def assign_to_cluster(
        X: NDArray,
        centroids: NDArray,
        n_clusters: int,
        p: float | int
    ) -> tuple[list[list[float]], list[int]]:
    clusters = [[] for _ in range(n_clusters)]
    labels = []

    for point in X:
        distances_to_each_centroid = pairwise_minkowski_distance(
            point, centroids, p)
        closest_centroid = int(np.argmin(distances_to_each_centroid))
        clusters[closest_centroid].append(point)
        labels.append(closest_centroid)
    return clusters, labels


# pylint: disable= too-few-public-methods, too-many-arguments
class LKMeans:
    def __init__(self,
                 n_clusters: int,
                 p: float | int = 2,
                 max_iter: int = 100,
                 max_iter_with_no_progress: int = 15) -> None:
        self.n_clusters = n_clusters
        self.max_iter = max_iter
        self.p = p
        self.max_iter_with_no_progress = max_iter_with_no_progress

        self.inertia_ = 0.
        self.cluster_centers_ = np.array([])

    @staticmethod
    def _init_centroids(data: NDArray, n_clusters: int) -> NDArray:
        indices = np.random.choice(
            data.shape[0], n_clusters, replace=False)
        centroids = data[indices]
        return centroids

    @staticmethod
    def _optimize_centroid(cluster: NDArray, p: float | int) -> NDArray:
        data_dimension = cluster.shape[1]
        new_centroid = np.array([])

        for coordinate_id in range(data_dimension):
            dimension_slice = cluster[:, coordinate_id]
            value = 0
            if p == 2:
                value = mean_optimizer(dimension_slice)
            if p == 1:
                value = median_optimizer(dimension_slice)
            elif 0 < p < 1:
                value = bound_optimizer(dimension_slice, p)
            elif p > 1:
                value = slsqp_optimizer(dimension_slice, p)
            else:
                raise ValueError('Parameter p must be greater than 0!')
            new_centroid = np.append(new_centroid, value)
        new_centroid = np.array(new_centroid)
        return new_centroid

    @staticmethod
    def _inertia(X: NDArray, centroids: NDArray) -> float:
        n_clusters = centroids.shape[0]
        distances = np.empty((X.shape[0], n_clusters))
        for i in range(n_clusters):
            distances[:, i] = np.sum((X - centroids[i, :])**2, axis=1)
        return np.sum(np.min(distances, axis=1))

    def fit(self, X: NDArray) -> None:
        centroids = self._init_centroids(X, self.n_clusters)

        iter_with_no_progress = 0
        for _ in range(self.max_iter):
            if iter_with_no_progress >= self.max_iter_with_no_progress:
                break

            bias_centroids = deepcopy(centroids)
            clusters, _ = assign_to_cluster(
                X, centroids, self.n_clusters, self.p)

            # update centroids using the specified optimizer
            for cluster_id, cluster in enumerate(clusters):
                cluster = np.array(cluster, copy=True)
                centroids[cluster_id] = deepcopy(
                    self._optimize_centroid(cluster, self.p)
                )

            if np.array_equal(bias_centroids, centroids):
                iter_with_no_progress += 1
            else:
                iter_with_no_progress = 0

        self.inertia_ = self._inertia(X, centroids)
        self.cluster_centers_ = deepcopy(centroids)

    def predict(self, X: NDArray) -> list[int]:
        _, labels = assign_to_cluster(
            X, self.cluster_centers_, self.n_clusters, self.p)
        return labels

    def fit_predict(self, X: NDArray) -> list[int]:
        self.fit(X)
        labels = self.predict(X)
        return labels



import numpy as np
from numpy.typing import NDArray

from lkmeans.examples.experiment import get_covariance_matrix


def get_experiment_data(num_clusters: int, dimension: int) -> tuple[int, float, list[NDArray], list[NDArray]]:
    '''
    Function for generation the synthetic data for experiments
    by dimension and number of clusters
    '''
    n_clusters: int = 0
    prob: float = 0.
    mu_prefix: list[list[float | int]] = [[]]
    sigma_list: list[float | int] = []

    if num_clusters == 2:
        print('Experiment with 2 clusters')
        n_clusters = 2
        sigma_list = [1, 1]
        prob = 0.5
        mu_prefix = [[-4, 0], [4, 0]]

    elif num_clusters == 3:
        print('Experiment with 3 clusters')
        n_clusters = 3
        sigma_list = [1, 1, 1]
        prob = 1/3
        mu_prefix = [[4, 0, 0], [0, 4, 0], [0, 0, 4]]

    else:
        raise KeyError(f'Not supported experiment type: {num_clusters}')

    mu_list = [np.array([x + [0] * (dimension - len(x))]) for x in mu_prefix]
    cov_matrix = [get_covariance_matrix(sigma, dimension) for sigma in sigma_list]
    return n_clusters, prob, mu_list, cov_matrix
